/*
								+---------------------------------+
								|                                 |
								|    ***   PADS-like menu  ***    |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|             Menu.cpp            |
								|                                 |
								+---------------------------------+
*/

#include "CallStack.h"
#include <windows.h>
#include <gl/gl.h>
#include <stdarg.h>
#include <stdio.h>
#include <vector>
#include <math.h>
#include <time.h>
#include <gl/glext.h>
#include <gl/wglext.h>
#include "OpenGLState.h"
#include "Vector.h"
#include "Pashik_TextParser2.h"
#include "StrokeFont.h"
#include "Menu.h"

#if defined(_MSC_VER) && !defined(__MWERKS__) && !defined(for)
#define for if(0) {} else for
#endif

/*
 *	char *p_s_Clone(const char *p_s_str)
 *		- alloc a new bufer and copy the contents
 */
char *p_s_Clone(const char *p_s_str)
{
	if(!p_s_str)
		return 0;

	char *p_s_tmp;
	if(!(p_s_tmp = new char[strlen(p_s_str) + 1]))
		return 0;
	strcpy(p_s_tmp, p_s_str);

	return p_s_tmp;
}

/*
 *								=== TToggleChoice ===
 */

TToggleChoice::TToggleChoice()
	:p_s_name(0), n_value(0)
{
}

/*
 *	bool TToggleChoice::Load(const /*const* / CTextParser &r_parser)
 *		- load from parser
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 *		- parser must be locked in node 'toggle_group',
 *		  it remains in it after return despite the result value
 */
bool TToggleChoice::Load(/*const*/ CTextParser &r_parser)
{
	Free();

	p_s_name = (char*)r_parser.p_s_Value("p_s_name", 0, false);
	if(!p_s_name || (!(p_s_name = p_s_Clone(p_s_name))))
		return false;
	n_value = r_parser.n_Value("n_value", -1, false);
	if(r_parser.n_GetLastError() != Parser_Ok)
		return false;
	// get the name and value - mandatory

	return true;
}

void TToggleChoice::Free()
{
	if(p_s_name) {
		delete[] p_s_name;
		p_s_name = 0;
	}
}

/*
 *								=== ~TToggleChoice ===
 */

/*
 *								=== TToggleGroup ===
 */

TToggleGroup::TToggleGroup()
	:n_id(-1)
{
}

/*
 *	bool TToggleGroup::Load(/*const* / CTextParser &r_parser)
 *		- load from parser
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 *		- parser must be locked in node 'toggle_group',
 *		  it remains in it after return despite the result value
 */
bool TToggleGroup::Load(/*const*/ CTextParser &r_parser)
{
	Free();

	n_id = r_parser.n_Value("n_id", -1, false);
	if(r_parser.n_GetLastError() != Parser_Ok)
		return false;
	// get id - mandatory

	if(r_parser.b_IsNode("choice")) {
		if(!r_parser.LockNode("choice"))
			return false;
		do {
			TToggleChoice t_choice;

			if(!t_choice.Load(r_parser)) {
				r_parser.LockParentNode();
				return false;
			}
			// parse

			choice_list.push_back(t_choice);
			if(!choice_list.size()) {
				t_choice.Free();
				r_parser.LockParentNode();
				return false;
			}
			// put into the list

			memset(&t_choice, 0, sizeof(TToggleChoice));
		} while(r_parser.LockNextNode());
		r_parser.LockParentNode();
	}
	// load choices

	_ASSERTE(r_parser.n_GetLastError() == Parser_Ok);

	return true;
}

/*
 *	const char *TToggleGroup::p_s_Translate(int n_value) const
 *		- translate integer value to string (return 0 if it doesn't have such a value)
 */
const char *TToggleGroup::p_s_Translate(int n_value) const
{
	for(const TToggleChoice *p_cur = choice_list.begin(); p_cur < choice_list.end(); p_cur ++) {
		if(p_cur->n_value == n_value)
			return p_cur->p_s_name;
	}
	return 0;
}

int TToggleGroup::n_PrevValue(int n_value) const
{
	for(const TToggleChoice *p_cur = choice_list.begin(); p_cur < choice_list.end(); p_cur ++) {
		if(p_cur->n_value == n_value) {
			if(p_cur - 1 >= choice_list.begin())
				return (p_cur - 1)->n_value;
			return p_cur->n_value;
		}
	}
	if(choice_list.size())
		return choice_list[0].n_value;
	return 0;
}

int TToggleGroup::n_NextValue(int n_value) const
{
	for(const TToggleChoice *p_cur = choice_list.begin(); p_cur < choice_list.end(); p_cur ++) {
		if(p_cur->n_value == n_value) {
			if(p_cur + 1 < choice_list.end())
				return (p_cur + 1)->n_value;
			return p_cur->n_value;
		}
	}
	if(choice_list.size())
		return choice_list[0].n_value;
	return 0;
}

int TToggleGroup::n_FirstValue(int n_value) const
{
	if(choice_list.size())
		return choice_list[0].n_value;
	return 0;
}

int TToggleGroup::n_LastValue(int n_value) const
{
	if(choice_list.size())
		return choice_list[choice_list.size() - 1].n_value;
	return 0;
}

int TToggleGroup::n_Value_Num() const
{
	return choice_list.size();
}

int TToggleGroup::n_Value_Index(int n_value) const
{
	for(const TToggleChoice *p_cur = choice_list.begin(); p_cur < choice_list.end(); p_cur ++) {
		if(p_cur->n_value == n_value)
			return p_cur - choice_list.begin();
	}
	return -1;
}

int TToggleGroup::n_Value(int n_index) const
{
	return choice_list[n_index].n_value;
}

void TToggleGroup::Free()
{
	for(TToggleChoice *p_cur = choice_list.begin(); p_cur < choice_list.end(); p_cur ++)
		p_cur->Free();
	choice_list.clear();
}

/*
 *								=== ~TToggleGroup ===
 */

/*
 *								=== TWatchInfo ===
 */

TWatchInfo::TWatchInfo()
	:p_s_name(0), p_n_value(0), p_f_value(0), p_s_value(0)
{
}

bool TWatchInfo::b_Initialized() const
{
	switch(n_data_type) {
	case watch_Int:
		return p_n_value != 0;
	case watch_Float:
		return p_f_value != 0;
	case watch_String:
		return p_s_value != 0;
	}
	return false;
}

/*
 *	bool TWatchInfo::Load(/*const* / CTextParser &r_parser)
 *		- load from parser
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 *		- parser must be locked in node 'watch',
 *		  it remains in it after return despite the result value
 */
bool TWatchInfo::Load(/*const*/ CTextParser &r_parser)
{
	Free();

	p_s_name = (char*)r_parser.p_s_Value("p_s_name", 0, false);
	if(!p_s_name || (!(p_s_name = p_s_Clone(p_s_name))))
		return false;
	n_data_type = r_parser.n_Value("n_type", -1, false);
	if(r_parser.n_GetLastError() != Parser_Ok)
		return false;
	// get the name and type - mandatory

	if(n_data_type < watch_Int || n_data_type > watch_String)
		return false;
	// type check

	_ASSERTE(r_parser.n_GetLastError() == Parser_Ok);

	return true;
}

void TWatchInfo::Free()
{
	if(p_s_name) {
		delete[] p_s_name;
		p_s_name = 0;
	}
	p_n_value = 0;
	p_f_value = 0;
	p_s_value = 0;
}

/*
 *								=== ~TWatchInfo ===
 */

/*
 *								=== TMenuButton ===
 */

TMenuButton::TMenuButton()
	:p_s_name(0), p_s_name_buffer(0), n_flags(0), n_name_length(0)
{
}

/*
 *	bool TMenuButton::Load(/*const* / CTextParser &r_parser)
 *		- load from parser
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 *		- parser must be locked in node 'button',
 *		  it remains in it after return despite the result value
 *		- all fields are voluntary
 *		- field sel must contain node watch and just one of nodes toggle / range / str_edit
 *		  data type of watch must correspond to presented node
 *		- range and str_edit nodes may be empty
 */
bool TMenuButton::Load(/*const*/ CTextParser &r_parser)
{
	Free();

	n_flags = 0;

	p_s_name = (char*)r_parser.p_s_Value("p_s_name", 0, false);
	if(p_s_name && (!(p_s_name = p_s_Clone(p_s_name))))
		return false;
	// read name

	n_preferred_index = r_parser.n_Value("n_index", -1, false);
	// read preferred index

	n_token = r_parser.n_Value("n_token", 0, false);
	// read token

	if((t_link.n_target = r_parser.n_Value("n_link_target", -1, false)) != -1)
		n_flags |= bf_Link;
	// read link target

	t_callback.p_s_name = (char*)r_parser.p_s_Value("p_s_callback_name", 0, false);
	if(t_callback.p_s_name && (!(t_callback.p_s_name = p_s_Clone(t_callback.p_s_name))))
		return false;
	else {
		n_flags |= bf_Callback;
		t_callback.p_handler = 0;
	}
	// read callback name

	r_parser.n_GetLastError(); // kill possible errors

	if(r_parser.b_IsNode("watch")) {
		if(!r_parser.LockNode("watch"))
			return false;
		if(!t_watch.Load(r_parser)) {
			r_parser.LockParentNode();
			return false;
		}
		if(!r_parser.LockParentNode())
			return false;
		n_flags |= bf_ShowWatch;
	}
	// read watch

	if(r_parser.b_IsNode("sel")) {
		if(!r_parser.LockNode("sel"))
			return false;
		if(!r_parser.LockNode("watch")) {
			r_parser.LockParentNode();
			return false;
		}
		if(!t_sel.t_sel_value.Load(r_parser)) {
			r_parser.LockParentNode();
			r_parser.LockParentNode();
			return false;
		}
		if(!r_parser.LockParentNode())
			return false;
		// read set / watch

		switch(t_sel.t_sel_value.n_data_type) {
		case watch_Int:
			if(!r_parser.b_IsNode("toggle"))
				return false;
			if(!r_parser.LockNode("toggle")) {
				r_parser.LockParentNode();
				return false;
			}
			t_sel.t_toggle.n_group_id = r_parser.n_Value("n_id", -1, false);
			if(r_parser.n_GetLastError() != Parser_Ok) {
				r_parser.LockParentNode();
				r_parser.LockParentNode();
				return false;
			}
			if(!r_parser.LockParentNode())
				return false;

			n_flags |= bf_Sel_ToggleVal;
			break;
		case watch_Float:
			if(!r_parser.b_IsNode("range"))
				return false;
			if(!r_parser.LockNode("range")) {
				r_parser.LockParentNode();
				return false;
			}
			t_sel.t_range.f_min = r_parser.f_Value("f_min", .0f, false);
			t_sel.t_range.f_max = r_parser.f_Value("f_max", 1.0f, false);
			t_sel.t_range.f_step = r_parser.f_Value("f_step", .0f, false);
			r_parser.n_GetLastError();
			if(!r_parser.LockParentNode())
				return false;

			t_sel.t_range.p_slider = 0;

			n_flags |= bf_Sel_RangeVal;
			break;
		case watch_String:
			if(!r_parser.b_IsNode("str_edit"))
				return false;
			if(!r_parser.LockNode("str_edit")) {
				r_parser.LockParentNode();
				return false;
			}
			t_sel.t_edit_str.n_max_length = r_parser.n_Value("n_max_length", -1, false);
			r_parser.n_GetLastError();
			if(!r_parser.LockParentNode())
				return false;

			t_sel.t_edit_str.p_edit_box = 0;

			n_flags |= bf_Sel_EditStr;
			break;
		default:
			r_parser.LockParentNode();
			// unlock from "set"
			return false;
		}
		// must have one of children

		if(!r_parser.LockParentNode())
			return false;
		// unlock from "set"
	}
	// read set

	_ASSERTE(r_parser.n_GetLastError() == Parser_Ok);

	return true;
}

static void _glDrawRect(CGLState *p_state, Vector2f v_min, Vector2f v_max,
	Vector3f v_color, float f_alpha, GLenum n_polygon_mode)
{
	p_state->Color4f(v_color.x, v_color.y, v_color.z, f_alpha);
	p_state->PolygonMode(GL_FRONT_AND_BACK, n_polygon_mode);

	glBegin(GL_QUADS);
	glVertex2f(v_min.x, v_min.y);
	glVertex2f(v_min.x, v_max.y);
	glVertex2f(v_max.x, v_max.y);
	glVertex2f(v_max.x, v_min.y);
	glEnd();
}

static void _glDrawText(CGLState *p_state, const CStrokeFont *p_font,
	Vector2f v_min, Vector2f v_max, Vector3f v_color, const char *p_s_str)
{
	Vector2f v_pos(((v_max.x - v_min.x) - p_font->f_StringWidth(p_s_str)) * .5f,
		((v_max.y - v_min.y) + p_font->v_CharacterOffset("\n", 1).y) * .5f);
	v_pos += v_min;
	// calc text position

	p_state->Color4f(v_color.x, v_color.y, v_color.z, 1.0f);

	p_state->EnableClipPlane(0);
	p_state->EnableClipPlane(1);
	p_state->EnableClipPlane(2);
	p_state->EnableClipPlane(3);
	//
	p_state->ClipPlane(0, 1, 0, 0, -v_min.x);
	p_state->ClipPlane(1, 0, 1, 0, -v_min.y);
	p_state->ClipPlane(2, -1, 0, 0, v_max.x);
	p_state->ClipPlane(3, 0, -1, 0, v_max.y);
	// setup clip-planes

	p_font->Plot_String(p_state, v_pos, p_s_str);
	// draw text

	p_state->DisableClipPlane(0);
	p_state->DisableClipPlane(1);
	p_state->DisableClipPlane(2);
	p_state->DisableClipPlane(3);
}

void TMenuButton::Draw(CGLState *p_state, const CStrokeFont *p_font, Vector2f v_min,
	Vector2f v_max, const CPadsMenu::TMenuColor &r_t_menu_color, bool b_highlight)
{
	p_state->EnableBlend();
	p_state->BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	_glDrawRect(p_state, v_min, v_max,
		(n_flags & bf_Sel_Mask && t_sel.b_active)? r_t_menu_color.v_edit_color :
		((b_highlight)? r_t_menu_color.v_focus_color : r_t_menu_color.v_bk_color),
		r_t_menu_color.f_alpha, GL_FILL);
	// fill

	_glDrawRect(p_state, v_min, v_max, r_t_menu_color.v_border_color, 1.0f, GL_LINE);
	// border

	const char *p_s_caption;
	if(n_flags & bf_ShowWatch && t_watch.b_Initialized()) {
		switch(t_watch.n_data_type) {
		case watch_Int:
			if(n_flags & bf_Sel_ToggleVal && t_watch.p_n_value == t_sel.t_sel_value.p_n_value &&
			   t_sel.t_toggle.p_toggle_group) {
				p_s_name_buffer = p_s_Format(p_s_name_buffer, n_name_length, p_s_name,
					t_sel.t_toggle.p_toggle_group->p_s_Translate(*t_watch.p_n_value));
			} else {
				p_s_name_buffer = p_s_Format(p_s_name_buffer, n_name_length,
					p_s_name, *t_watch.p_n_value);
			}
			break;
		case watch_Float:
			p_s_name_buffer = p_s_Format(p_s_name_buffer, n_name_length,
				p_s_name, *t_watch.p_f_value);
			break;
		case watch_String:
			p_s_name_buffer = p_s_Format(p_s_name_buffer, n_name_length,
				p_s_name, t_watch.p_s_value->c_str());
			break;
		}

		p_s_caption = p_s_name_buffer;
	} else
		p_s_caption = p_s_name;
	// generate button caption

	Vector2f v_pos(p_font->f_Zoom() * 3, ((v_max.y - v_min.y) + p_font->v_CharacterOffset("\n", 1).y) * .5f);
	v_pos += v_min;
	// calc text position

	p_state->Color3f(r_t_menu_color.v_font_color.x, r_t_menu_color.v_font_color.y,
		r_t_menu_color.v_font_color.z);

	p_state->EnableClipPlane(0);
	p_state->EnableClipPlane(1);
	p_state->EnableClipPlane(2);
	p_state->EnableClipPlane(3);
	//
	p_state->ClipPlane(0, 1, 0, 0, -v_min.x);
	p_state->ClipPlane(1, 0, 1, 0, -v_min.y);
	p_state->ClipPlane(2, -1, 0, 0, v_max.x);
	p_state->ClipPlane(3, 0, -1, 0, v_max.y);
	// setup clip-planes

	p_font->Plot_String(p_state, v_pos, (p_s_caption)? p_s_caption : "");
	// draw text

	p_state->DisableClipPlane(0);
	p_state->DisableClipPlane(1);
	p_state->DisableClipPlane(2);
	p_state->DisableClipPlane(3);

	if(n_flags & bf_Sel_Mask && t_sel.b_active) {
		switch(t_sel.t_sel_value.n_data_type) {
		case watch_Int:
			{
				Vector2f v_size = v_max - v_min;

				float f_button_height = v_size.y;

				v_size.x += v_size.y;
				v_size.y = 0;

				float f_height = t_sel.t_toggle.p_toggle_group->n_Value_Num() * f_button_height;

				float f_offset = t_sel.t_toggle.p_toggle_group->n_Value_Index(
					*t_sel.t_sel_value.p_n_value) * f_button_height;

				Vector2f v_max2 = v_max + Vector2f(0, f_offset) + v_size;
				Vector2f v_min2 = v_min + Vector2f(0, f_offset + f_button_height - f_height) + v_size;
				
				_glDrawRect(p_state, v_min2, v_max2,
					r_t_menu_color.v_bk_color, r_t_menu_color.f_alpha, GL_FILL);
				// fill

				_glDrawRect(p_state, v_min2, v_max2, r_t_menu_color.v_border_color, 1.0f, GL_LINE);
				// border

				v_min2 = v_max2 + v_min - v_max;
				// to size of a single button

				for(int i = 0; i < t_sel.t_toggle.p_toggle_group->n_Value_Num(); i ++) {
					_glDrawText(p_state, p_font, v_min2, v_max2, r_t_menu_color.v_font_color,
						t_sel.t_toggle.p_toggle_group->p_s_Translate(
						t_sel.t_toggle.p_toggle_group->n_Value(i)));

					v_min2.y -= f_button_height;
					v_max2.y -= f_button_height;
				}
				// label captions
			}
			break;
		case watch_Float:
			if(t_sel.t_range.p_slider) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_range.p_slider->Draw(p_state, v_min + v_size,
					v_max + v_size, r_t_menu_color);
			}
			break;
		case watch_String:
			if(t_sel.t_edit_str.p_edit_box) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_edit_str.p_edit_box->Draw(p_state, p_font, v_min + v_size,
					v_max + v_size, r_t_menu_color);
			}
			break;
		}
	}
}

bool TMenuButton::b_IsEdited() const
{
	return n_flags & bf_Sel_Mask && t_sel.b_active;
}

int TMenuButton::n_OnMouseMove(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
	if(n_flags & bf_Sel_Mask && t_sel.b_active) {
		switch(t_sel.t_sel_value.n_data_type) {
		case watch_Int:
			break;
		case watch_Float:
			if(t_sel.t_range.p_slider) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_range.p_slider->OnMouseMove(v_pos, v_min + v_size, v_max + v_size);
			}
			break;
		case watch_String:
			if(t_sel.t_edit_str.p_edit_box) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_edit_str.p_edit_box->OnMouseMove(v_pos, v_min + v_size, v_max + v_size);
			}
			break;
		}
	}

	return -1;
}

int TMenuButton::n_OnMouseDown(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
	if(!t_sel.b_active)
		return n_OnKeyPress(VK_RETURN, 0);

	if(n_flags & bf_Sel_Mask && t_sel.b_active) {
		switch(t_sel.t_sel_value.n_data_type) {
		case watch_Int:
			{
				Vector2f v_size = v_max - v_min;
				float f_button_height = v_size.y;
				v_size.x += v_size.y;
				v_size.y = 0;
				float f_height = t_sel.t_toggle.p_toggle_group->n_Value_Num() * f_button_height;
				float f_offset = t_sel.t_toggle.p_toggle_group->n_Value_Index(
					*t_sel.t_sel_value.p_n_value) * f_button_height;
				Vector2f v_max2 = v_max + Vector2f(0, f_offset) + v_size;
				Vector2f v_min2 = v_max2 + v_min - v_max;
				// calc vectors

				if(v_pos.x >= v_min2.x && v_pos.x <= v_max2.x) {
					for(int i = 0; i < t_sel.t_toggle.p_toggle_group->n_Value_Num(); i ++) {
						if(v_pos.y >= v_min2.y && v_pos.y <= v_max2.y) {
							*t_sel.t_sel_value.p_n_value = t_sel.t_toggle.p_toggle_group->n_Value(i);
							break;
						}

						v_min2.y -= f_button_height;
						v_max2.y -= f_button_height;
					}
				}
				// label captions

				return n_OnKeyPress(VK_RETURN, 0);
				// deactivate
			}
			break;
		case watch_Float:
			if(t_sel.t_range.p_slider) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_range.p_slider->OnMouseDown(v_pos, v_min + v_size, v_max + v_size);
			}
			break;
		case watch_String:
			if(t_sel.t_edit_str.p_edit_box) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_edit_str.p_edit_box->OnMouseMove(v_pos, v_min + v_size, v_max + v_size);
			}
			break;
		}
	}

	return -1;
}

int TMenuButton::n_OnMouseUp(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
	if(n_flags & bf_Sel_Mask && t_sel.b_active) {
		switch(t_sel.t_sel_value.n_data_type) {
		case watch_Float:
			if(t_sel.t_range.p_slider) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_range.p_slider->OnMouseUp(v_pos, v_min + v_size, v_max + v_size);
			}
			break;
		case watch_String:
			if(t_sel.t_edit_str.p_edit_box) {
				Vector2f v_size = v_max - v_min;
				v_size.x += v_size.y;
				v_size.y = 0;

				t_sel.t_edit_str.p_edit_box->OnMouseMove(v_pos, v_min + v_size, v_max + v_size);
			}
			break;
		}
	}

	return -1;
}

int TMenuButton::n_OnKeyPress(int n_code, char n_char)
{
	if(n_flags & bf_Sel_Mask && n_code == VK_RETURN && t_sel.t_sel_value.b_Initialized()) {
		// can edit ...

		if(!t_sel.b_active && DoCallback(event_EditBegin)) {
			if(t_sel.t_sel_value.n_data_type == watch_String) {
				if(!(t_sel.t_edit_str.p_edit_box = new CEditBox(t_sel.t_sel_value.p_s_value->c_str(),
				   0, t_sel.t_edit_str.n_max_length)))
					return -1;
				// alloc

				*t_sel.t_sel_value.p_s_value = t_sel.t_edit_str.p_edit_box->p_s_Text();
				// set

				DoCallback(event_EditChange);
				// let them know we changed the value (by allocating it)
			} else if(t_sel.t_sel_value.n_data_type == watch_Float) {
				if(!(t_sel.t_range.p_slider = new CSlider(t_sel.t_sel_value.p_f_value,
				   t_sel.t_range.f_min, t_sel.t_range.f_max, t_sel.t_range.f_step)))
					return -1;
				// create a new slider object
			}
			t_sel.b_active = true;

			return -1;
		} else if(t_sel.b_active && DoCallback(event_EditEnd)) {
			t_sel.b_active = false;
			if(t_sel.t_sel_value.n_data_type == watch_Float) {
				if(t_sel.t_range.p_slider) {
					delete t_sel.t_range.p_slider;
					t_sel.t_range.p_slider = 0;
				}
			} else if(t_sel.t_sel_value.n_data_type == watch_String) {
				if(t_sel.t_edit_str.p_edit_box->b_Dirty()) {
					*t_sel.t_sel_value.p_s_value = t_sel.t_edit_str.p_edit_box->p_s_Text();
					t_sel.t_edit_str.p_edit_box->ResetDirtyFlag();

					DoCallback(event_EditChange);
					// let them know we changed the value
				}
				// set

				if(t_sel.t_edit_str.p_edit_box) {
					delete t_sel.t_edit_str.p_edit_box;
					t_sel.t_edit_str.p_edit_box = 0;
				}
			}
		}
		// toggle editing
	}

	if(n_code == VK_RETURN && DoCallback(event_ButtonActivate)) {
		if(n_flags & bf_Link)
			return t_link.n_target;
	}
	// jump to it's target (note it happens *after* editing is finished if it's editable button)

	if(t_sel.b_active) {
		switch(t_sel.t_sel_value.n_data_type) {
		case watch_Int:
			if(n_code == VK_DOWN) {
				*t_sel.t_sel_value.p_n_value = t_sel.t_toggle.p_toggle_group->n_PrevValue(
					*t_sel.t_sel_value.p_n_value);
			} else if(n_code == VK_UP) {
				*t_sel.t_sel_value.p_n_value = t_sel.t_toggle.p_toggle_group->n_NextValue(
					*t_sel.t_sel_value.p_n_value);
			}
			break;
		case watch_Float:
			if(t_sel.t_range.p_slider)
				t_sel.t_range.p_slider->OnKeyPress(n_code, n_char);
			break;
		case watch_String:
			if(t_sel.t_edit_str.p_edit_box) {
				t_sel.t_edit_str.p_edit_box->OnKeyPress(n_code, n_char);
				if(t_sel.t_edit_str.p_edit_box->b_Dirty()) {
					*t_sel.t_sel_value.p_s_value = t_sel.t_edit_str.p_edit_box->p_s_Text();
					t_sel.t_edit_str.p_edit_box->ResetDirtyFlag();

					DoCallback(event_EditChange);
					// let them know we changed the value
				}
			}
			break;
		}
	}

	return -1; // we ain't going nowhere
}

bool TMenuButton::Find_ToggleGroup(const std::vector<TToggleGroup> &r_enum_list)
{
	if(n_flags & bf_Sel_ToggleVal && t_sel.t_sel_value.n_data_type == watch_Int) {
		for(const TToggleGroup *p_group = r_enum_list.begin(); p_group < r_enum_list.end(); p_group ++) {
			if(p_group->n_id == t_sel.t_toggle.n_group_id) {
				t_sel.t_toggle.p_toggle_group = p_group;
				return true;
			}
		}
		t_sel.t_toggle.p_toggle_group = 0;
		return false;
	}
	return true;
}

int TMenuButton::n_RegisterCallback(TCallbackHandler *p_callback_handler)
{
	if(!(n_flags & bf_Callback))
		return 0;
	if(!p_callback_handler->b_Handle(t_callback.p_s_name))
		return 0;
	t_callback.p_handler = p_callback_handler;
	return 1;
}

void TMenuButton::RemoveCallback(TCallbackHandler *p_callback_handler)
{
	if(n_flags & bf_Callback && t_callback.p_handler == p_callback_handler)
		t_callback.p_handler = 0;
}

int TMenuButton::n_AddWatch(const char *p_s_name, int *p_value_ptr)
{
	int n_result = 0;

	if(n_flags & bf_ShowWatch) {
		if(!strcmp(p_s_name, t_watch.p_s_name) && t_watch.n_data_type == watch_Int) {
			t_watch.p_n_value = p_value_ptr;
			n_result = 1;
		}
	}
	if(n_flags & bf_Sel_Mask) {
		if(!strcmp(p_s_name, t_sel.t_sel_value.p_s_name) &&
		   t_sel.t_sel_value.n_data_type == watch_Int) {
			t_sel.t_sel_value.p_n_value = p_value_ptr;
			n_result = 1;
		}
	}

	return n_result;
}

int TMenuButton::n_AddWatch(const char *p_s_name, float *p_value_ptr)
{
	int n_result = 0;

	if(n_flags & bf_ShowWatch) {
		if(!strcmp(p_s_name, t_watch.p_s_name) && t_watch.n_data_type == watch_Float) {
			t_watch.p_f_value = p_value_ptr;
			n_result = 1;
		}
	}
	if(n_flags & bf_Sel_Mask) {
		if(!strcmp(p_s_name, t_sel.t_sel_value.p_s_name) &&
		   t_sel.t_sel_value.n_data_type == watch_Float) {
			t_sel.t_sel_value.p_f_value = p_value_ptr;
			n_result = 1;
		}
	}

	return n_result;
}

int TMenuButton::n_AddWatch(const char *p_s_name, std::string *p_value_ptr)
{
	int n_result = 0;

	if(n_flags & bf_ShowWatch) {
		if(!strcmp(p_s_name, t_watch.p_s_name) && t_watch.n_data_type == watch_String) {
			t_watch.p_s_value = p_value_ptr;
			n_result = 1;
		}
	}
	if(n_flags & bf_Sel_Mask) {
		if(!strcmp(p_s_name, t_sel.t_sel_value.p_s_name) &&
		   t_sel.t_sel_value.n_data_type == watch_String) {
			t_sel.t_sel_value.p_s_value = p_value_ptr;
			n_result = 1;
		}
	}

	return n_result;
}

void TMenuButton::RemoveWatch(const char *p_s_name)
{
	if(n_flags & bf_ShowWatch) {
		if(!strcmp(p_s_name, t_watch.p_s_name)) {
			t_watch.p_n_value = 0;
			t_watch.p_f_value = 0;
			t_watch.p_s_value = 0;
		}
	}
	if(n_flags & bf_Sel_Mask) {
		if(!strcmp(p_s_name, t_sel.t_sel_value.p_s_name)) {
			t_sel.t_sel_value.p_n_value = 0;
			t_sel.t_sel_value.p_f_value = 0;
			t_sel.t_sel_value.p_s_value = 0;
		}
	}
}

bool TMenuButton::DoCallback(int n_event)
{
	if(!(n_flags & bf_Callback))
		return true;
	// we will pretend success

	return true; // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

	if(!t_callback.p_handler)
		return false;
	// uneregistered handler

	return t_callback.p_handler->operator ()(t_callback.p_s_name, this, n_event);
}

void TMenuButton::Free()
{
	if(p_s_name) {
		delete[] p_s_name;
		p_s_name = 0;
	}
	if(n_flags & bf_Callback && t_callback.p_s_name) {
		delete[] t_callback.p_s_name;
		t_callback.p_s_name = 0;
	}
	if(n_flags & bf_ShowWatch) {
		if(p_s_name_buffer) {
			delete[] p_s_name_buffer;
			p_s_name_buffer = 0;
			n_name_length = 0;
		}
		if(t_watch.p_s_name) {
			delete[] t_watch.p_s_name;
			t_watch.p_s_name = 0;
		}
	}
	if(n_flags & bf_Sel_Mask) {
		if(t_sel.t_sel_value.p_s_name) {
			delete[] t_sel.t_sel_value.p_s_name;
			t_sel.t_sel_value.p_s_name = 0;
		}
		if(n_flags & bf_Sel_EditStr && t_sel.t_edit_str.p_edit_box) {
			delete t_sel.t_edit_str.p_edit_box;
			t_sel.t_edit_str.p_edit_box = 0;
		}
	}
	if(n_flags & bf_Sel_RangeVal && t_sel.t_range.p_slider) {
		delete t_sel.t_range.p_slider;
		t_sel.t_range.p_slider = 0;
	}
}

/*
 *								=== ~TMenuButton ===
 */

/*
 *								=== TMenuPage ===
 */

TMenuPage::TMenuPage()
	:p_s_name(0), p_edit_button(0), n_active_button(0)
{
}

/*
 *	bool TMenuPage::Load(/*const* / CTextParser &r_parser)
 *		- load from parser
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 *		- parser must be locked in node 'menu',
 *		  it remains in it after return despite the result value
 */
bool TMenuPage::Load(/*const*/ CTextParser &r_parser)
{
	Free();

	p_s_name = (char*)r_parser.p_s_Value("p_s_name", 0, false);
	if(!p_s_name || (!(p_s_name = p_s_Clone(p_s_name))))
		return false;
	n_id = r_parser.n_Value("n_id", -1, false);
	if(r_parser.n_GetLastError() != Parser_Ok)
		return false;
	// get the name and id - mandatory

	if(r_parser.b_IsNode("button")) {
		if(!r_parser.LockNode("button"))
			return false;
		int n_index = 0;
		do {
			TMenuButton t_button;

			if(!t_button.Load(r_parser)) {
				r_parser.LockParentNode();
				return false;
			}
			// load

			if(t_button.n_PreferredIndex() == -1)
				t_button.SetIndex(n_index ++);
			else if(t_button.n_PreferredIndex() >= n_index)
				n_index = t_button.n_PreferredIndex() + 1;
			else
				t_button.SetIndex(n_index ++);
			// manage indices

			button_list.push_back(t_button);
			if(!button_list.size()) {
				t_button.Free();
				r_parser.LockParentNode();
				return false;
			}
			// add to list

			memset(&t_button, 0, sizeof(TMenuButton));
		} while(r_parser.LockNextNode());
		if(!r_parser.LockParentNode())
			return false;
	}
	// load button(s)

	_ASSERTE(r_parser.n_GetLastError() == Parser_Ok);

	return true;
}

Vector2f TMenuPage::v_ButtonMin(const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	int n_button_num, int n_button_index) const
{
	Vector2f v_min = r_t_menu_metrics.v_menu_offset +
		Vector2f(r_t_menu_metrics.v_button_spacing.x,
		(r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height) *
		(n_button_num - n_button_index) - r_t_menu_metrics.f_button_height);

	return v_min;
	// calc rectangle of first
}

Vector2f TMenuPage::v_ButtonMax(const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	int n_button_num, int n_button_index) const
{
	Vector2f v_max = r_t_menu_metrics.v_menu_offset +
		Vector2f(r_t_menu_metrics.f_width - r_t_menu_metrics.v_button_spacing.x,
		(r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height) *
		(n_button_num - n_button_index));

	return v_max;
	// calc rectangle of first
}

int TMenuPage::n_ButtonIndex(const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	int n_button_num, Vector2f v_mouse) const
{
	if(v_mouse.x < (r_t_menu_metrics.v_menu_offset.x + r_t_menu_metrics.v_button_spacing.x) ||
	   v_mouse.x > (r_t_menu_metrics.v_menu_offset.x + r_t_menu_metrics.f_width -
	   r_t_menu_metrics.v_button_spacing.x))
		return -1;
	// lies outside of menu column

	float f_min_y = v_ButtonMin(r_t_menu_metrics, n_button_num, 0).y;
	float f_max_y = v_ButtonMax(r_t_menu_metrics, n_button_num, 0).y;
	float f_height = r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
	for(int i = 0; i < n_button_num; i ++) {
		if(v_mouse.y >= f_min_y && v_mouse.y <= f_max_y)
			return i;

		f_min_y -= f_height;
		f_max_y -= f_height;
	}
	// see if mouse lies inside region of some button

	return -1;
}

bool TMenuPage::Find_ToggleGroups(const std::vector<TToggleGroup> &r_enum_list)
{
	bool b_result = true;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		b_result &= p_button->Find_ToggleGroup(r_enum_list);
	return b_result;
}

int TMenuPage::n_RegisterCallback(TCallbackHandler *p_callback_handler)
{
	int n_result = 0;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		n_result += p_button->n_RegisterCallback(p_callback_handler);
	return n_result;
}

void TMenuPage::RemoveCallback(TCallbackHandler *p_callback_handler)
{
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		p_button->RemoveCallback(p_callback_handler);
}

int TMenuPage::n_AddWatch(const char *p_s_name, int *p_value_ptr)
{
	int n_result = 0;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		n_result += p_button->n_AddWatch(p_s_name, p_value_ptr);
	return n_result;
}

int TMenuPage::n_AddWatch(const char *p_s_name, float *p_value_ptr)
{
	int n_result = 0;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		n_result += p_button->n_AddWatch(p_s_name, p_value_ptr);
	return n_result;
}

int TMenuPage::n_AddWatch(const char *p_s_name, std::string *p_value_ptr)
{
	int n_result = 0;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		n_result += p_button->n_AddWatch(p_s_name, p_value_ptr);
	return n_result;
}

void TMenuPage::RemoveWatch(const char *p_s_name)
{
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		p_button->RemoveWatch(p_s_name);
}

int TMenuPage::n_OnMouseMove(Vector2f v_pos, const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	int n_button_num)
{
	if(p_edit_button) {
		int n_result = p_edit_button->n_OnMouseMove(v_pos,
			v_ButtonMin(r_t_menu_metrics, n_button_num, n_active_button),
			v_ButtonMax(r_t_menu_metrics, n_button_num, n_active_button));
		if(!p_edit_button->b_IsEdited())
			p_edit_button = 0;
		return n_result;
	}

	int n_button = n_ButtonIndex(r_t_menu_metrics, n_button_num, v_pos);

	if(n_button != -1) {
		n_active_button = n_button;
		TMenuButton *p_button;
		if(!(p_button = p_Button(n_active_button)))
			return -1;
		int n_result = p_button/*button_list[n_active_button]*/->n_OnMouseMove(v_pos,
			v_ButtonMin(r_t_menu_metrics, n_button_num, n_active_button),
			v_ButtonMax(r_t_menu_metrics, n_button_num, n_active_button));
		if(p_button/*button_list[n_active_button]*/->b_IsEdited())
			p_edit_button = p_button/*button_list[n_active_button]*/;
		return n_result;
	}
	return -1;
}

TMenuButton *TMenuPage::p_Button(int n_index)
{
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++) {
		if(p_button->n_PreferredIndex() == n_index)
			return p_button;
	}
	return 0;
}

int TMenuPage::n_OnMouseDown(Vector2f v_pos, const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	int n_button_num)
{
	if(p_edit_button) {
		int n_result = p_edit_button->n_OnMouseDown(v_pos,
			v_ButtonMin(r_t_menu_metrics, n_button_num, n_active_button),
			v_ButtonMax(r_t_menu_metrics, n_button_num, n_active_button));
		if(!p_edit_button->b_IsEdited())
			p_edit_button = 0;
		return n_result;
	}
	
	/*int n_button = n_ButtonIndex(r_t_menu_metrics, n_button_num, v_pos);

	if(n_button != -1) {
		n_active_button = n_button;*/
		TMenuButton *p_button;
		if(!(p_button = p_Button(n_active_button)))
			return -1;
		int n_result = p_button/*button_list[n_active_button]*/->n_OnMouseDown(v_pos,
			v_ButtonMin(r_t_menu_metrics, n_button_num, n_active_button),
			v_ButtonMax(r_t_menu_metrics, n_button_num, n_active_button));
		if(p_button/*button_list[n_active_button]*/->b_IsEdited())
			p_edit_button = p_button/*button_list[n_active_button]*/;
		return n_result;
	/*}

	return -1;*/
}

int TMenuPage::n_OnMouseUp(Vector2f v_pos, const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	int n_button_num)
{
	if(p_edit_button) {
		int n_result = p_edit_button->n_OnMouseUp(v_pos,
			v_ButtonMin(r_t_menu_metrics, n_button_num, n_active_button),
			v_ButtonMax(r_t_menu_metrics, n_button_num, n_active_button));
		if(!p_edit_button->b_IsEdited())
			p_edit_button = 0;
		return n_result;
	}

	/*int n_button = n_ButtonIndex(r_t_menu_metrics, n_button_num, v_pos);

	if(n_button != -1) {
		n_active_button = n_button;*/
		TMenuButton *p_button;
		if(!(p_button = p_Button(n_active_button)))
			return -1;
		int n_result = p_button/*button_list[n_active_button]*/->n_OnMouseUp(v_pos,
			v_ButtonMin(r_t_menu_metrics, n_button_num, n_active_button),
			v_ButtonMax(r_t_menu_metrics, n_button_num, n_active_button));
		if(p_button/*button_list[n_active_button]*/->b_IsEdited())
			p_edit_button = p_button/*button_list[n_active_button]*/;
		return n_result;
	/*}

	return -1;*/
}

int TMenuPage::n_OnKeyPress(int n_code, int n_button_num)
{
	if(p_edit_button) {
		char n_char = 0;

		unsigned short p_char[2];
		unsigned char p_key_states[256] = {0};
		p_key_states[VK_CAPITAL] = GetKeyState(VK_CAPITAL);
		p_key_states[VK_SHIFT] = GetKeyState(VK_SHIFT);
		p_key_states[VK_CONTROL] = GetKeyState(VK_CONTROL);
		p_key_states[VK_MENU] = GetKeyState(VK_MENU);
		if(ToAscii(n_code, n_code, p_key_states, p_char, false) == 0)
			n_char = 0;
		else
			n_char = p_char[0];
		// win-based translation

		/*if(n_code >= 'A' && n_code <= 'Z') {
			n_char = n_code +
				((GetAsyncKeyState(VK_SHIFT) & 0xff00)? 0 : 'a' - 'A');
		} else if(n_code >= VK_NUMPAD0 && n_code <= VK_NUMPAD9)
			n_char = n_code - VK_NUMPAD0 + '0';
		else if(n_code >= '0' && n_code <= '9')
			n_char = n_code;
		else if((n_code >= '!' && n_code <= '/') || // !"#$%&'()*+,-./
		   (n_code >= ':' && n_code <= '@') || // :;<=>?@
		   (n_code >= '[' && n_code <= '`') || // [\]^_`
		   (n_code >= '{' && n_code <= '~')) // {|}~
			n_char = n_code;*/
		// try to convert to displayable characters

		int n_result = p_edit_button->n_OnKeyPress(n_code, n_char);
		// send the message to the propper button

		if(!p_edit_button->b_IsEdited())
			p_edit_button = 0;
		// see if we actually finished editation

		return n_result;
	} else {
		switch(n_code) {
		case VK_UP:
		case VK_DOWN:
			n_active_button = (n_active_button + 1 - 2 *
				(n_code == VK_UP) + n_button_num) % n_button_num;
			break;

		case VK_RETURN:
		default:
			{
				int n_key = -1;
				if(n_code == VK_RETURN)
					n_key = n_active_button;
				else if(n_code >= '0' && n_code <= '9')
					n_key = n_code - '0';
				else if(n_code >= VK_NUMPAD0 && n_code <= VK_NUMPAD9)
					n_key = n_code - VK_NUMPAD0;
				else if(n_code >= 'A' && n_code <= 'Z')
					n_key = n_code - 'A' + 10;
				else if(n_code >= 'a' && n_code <= 'z')
					n_key = n_code - 'a' + 10;
				if(n_key >= 0 && n_key < n_button_num) {
					n_active_button = n_key;
					TMenuButton *p_button;
					if(!(p_button = p_Button(n_active_button)))
						return -1;
					int n_result = p_button/*button_list[n_active_button]*/->n_OnKeyPress(VK_RETURN, 0);
					if(p_button/*button_list[n_active_button]*/->b_IsEdited())
						p_edit_button = p_button/*button_list[n_active_button]*/;
					return n_result;
				}
			}
			// pads-like behavior where you select keys
			break;
		}
	}

	return -1;
}

void TMenuPage::Draw(CGLState *p_state, const CStrokeFont *p_font,
	const CPadsMenu::TMenuMetrics &r_t_menu_metrics,
	const CPadsMenu::TMenuColor &r_t_menu_color, int n_button_num)
{
	p_state->BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	p_state->EnableBlend();

	Vector2f v_min = r_t_menu_metrics.v_menu_offset;
	Vector2f v_max = v_min + Vector2f(r_t_menu_metrics.f_width,
		r_t_menu_metrics.v_button_spacing.y +
		(r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height) * n_button_num);
	// calc rectangle of the whole menu, later on we will be filling it up with buttons

	_glDrawRect(p_state, v_max - Vector2f(r_t_menu_metrics.f_width, 0),
		v_max + Vector2f(0, r_t_menu_metrics.f_title_height),
		r_t_menu_color.v_bk_color, r_t_menu_color.f_alpha, GL_FILL);
	// title fill

	_glDrawRect(p_state, v_max - Vector2f(r_t_menu_metrics.f_width, 0),
		v_max + Vector2f(0, r_t_menu_metrics.f_title_height),
		r_t_menu_color.v_border_color, 1.0f, GL_LINE);
	// title border

	_glDrawRect(p_state, v_min, v_max, r_t_menu_color.v_border_color, 1.0f, GL_LINE);
	// whole border

	_glDrawText(p_state, p_font, v_max - Vector2f(r_t_menu_metrics.f_width, 0),
		v_max + Vector2f(0, r_t_menu_metrics.f_title_height),
		r_t_menu_color.v_font_color, p_s_name);
	// menu page title

	Vector2f v_button_max = v_max - r_t_menu_metrics.v_button_spacing;
	Vector2f v_button_min = Vector2f(v_min.x + r_t_menu_metrics.v_button_spacing.x,
		v_button_max.y - r_t_menu_metrics.f_button_height);
	// calc rectangle of first button

	int n_index = 0;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++) {
		for(; n_index < p_button->n_PreferredIndex(); n_index ++) {
			_glDrawRect(p_state, v_button_min, v_button_max,
				(n_active_button == n_index)? r_t_menu_color.v_focus_color :
				r_t_menu_color.v_bk_color, r_t_menu_color.f_alpha, GL_FILL);
			// fake button fill

			_glDrawRect(p_state, v_button_min, v_button_max,
				r_t_menu_color.v_border_color, 1.0f, GL_LINE);
			// fake button border

			v_button_max.y -= r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
			v_button_min.y -= r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
			// calculate next button's rectangle
		}

		v_min = v_ButtonMin(r_t_menu_metrics, n_button_num, n_index);
		v_max = v_ButtonMax(r_t_menu_metrics, n_button_num, n_index);

		p_button->Draw(p_state, p_font, v_button_min,
			v_button_max, r_t_menu_color, n_index == n_active_button && !p_edit_button);
		// draw button

		n_index ++;

		v_button_max.y -= r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
		v_button_min.y -= r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
		// calculate next button's rectangle
	}

	for(; n_index < n_button_num; n_index ++) {
		_glDrawRect(p_state, v_button_min, v_button_max,
			(n_active_button == n_index)? r_t_menu_color.v_focus_color :
			r_t_menu_color.v_bk_color, r_t_menu_color.f_alpha, GL_FILL);
		// fake button fill

		_glDrawRect(p_state, v_button_min, v_button_max,
			r_t_menu_color.v_border_color, 1.0f, GL_LINE);
		// fake button border

		v_button_max.y -= r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
		v_button_min.y -= r_t_menu_metrics.v_button_spacing.y + r_t_menu_metrics.f_button_height;
		// calculate next button's rectangle
	}
	// draw rest of fake buttons at the bottom
}

void TMenuPage::Free()
{
	if(p_s_name) {
		delete[] p_s_name;
		p_s_name = 0;
	}
	p_edit_button = 0;
	for(TMenuButton *p_button = button_list.begin(); p_button < button_list.end(); p_button ++)
		p_button->Free();
	button_list.clear();
}

/*
 *								=== ~TMenuPage ===
 */

/*
 *								=== CPadsMenu ===
 */

CPadsMenu::CPadsMenu()
	:m_p_stroke_font(0), m_p_cur_page(0)
{
}

CPadsMenu::~CPadsMenu()
{
	Free();
}

/*
 *	bool CPadsMenu::Load(const char *p_s_filename)
 *		- load from file p_s_filename
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 */
bool CPadsMenu::Load(const char *p_s_filename)
{
	CTextParser parser;

	if(!parser.Attach(p_s_filename) || !parser.LoadFromFile())
		return false;

	bool b_result = Load(parser);

	parser.Detach();

	return b_result;
}

/*
 *	bool CPadsMenu::Load(/*const* / CTextParser &r_parser)
 *		- load from parser
 *		- free any previously-allocated data
 *		- return true on success, false on parse error or not enough memory
 */
bool CPadsMenu::Load(/*const*/ CTextParser &r_parser)
{
	Free();

	if(r_parser.b_IsNode("toggle_group")) {
		if(!r_parser.LockNode("toggle_group"))
			return false;
		do {
			TToggleGroup t_group;

			if(!t_group.Load(r_parser)) {
				r_parser.LockParentNode();
				return false;
			}
			// parse

			m_enum_list.push_back(t_group);
			if(!m_enum_list.size()) {
				t_group.Free();
				r_parser.LockParentNode();
				return false;
			}
			// put into the list

			memset(&t_group, 0, sizeof(TToggleGroup));
		} while(r_parser.LockNextNode());
		if(!r_parser.LockParentNode())
			return false;
	}
	// load combo-box enum sets

	if(!r_parser.LockNode("menu"))
		return false;

	m_n_first_page_id = r_parser.n_Value("n_first_page_id", -1, false);
	m_n_button_num = r_parser.n_Value("n_button_num", 10, false);

	float f_font_size = r_parser.f_Value("f_font_size", .0375f, false);
	const char *p_s_font_file = r_parser.p_s_Value("p_s_font", 0, false);

	if(r_parser.n_GetLastError() != Parser_Ok || !p_s_font_file ||
	   !m_t_color.Load(r_parser) || !m_t_metrics.Load(r_parser)) {
		r_parser.LockParentNode();
		return false;
	}
	// load metrics / color scheme

	if(!(m_p_stroke_font = new CStrokeFont()) ||
	   !m_p_stroke_font->Load(p_s_font_file, 0)) {
		r_parser.LockParentNode();
		return false;
	}
	m_p_stroke_font->Zoom(f_font_size);
	// load font

	if(r_parser.b_IsNode("page")) {
		if(!r_parser.LockNode("page"))
			return false;
		do {
			TMenuPage *p_page;

			if(!(p_page = new TMenuPage())) {
				r_parser.LockParentNode();
				r_parser.LockParentNode();
				return false;
			}
			// alloc

			if(!p_page->Load(r_parser) || !p_page->Find_ToggleGroups(m_enum_list)) {
				p_page->Free();
				delete p_page;
				r_parser.LockParentNode();
				r_parser.LockParentNode();
				return false;
			}
			// parse

			m_page_list.push_back(p_page);
			if(!m_page_list.size()) {
				p_page->Free();
				delete p_page;
				r_parser.LockParentNode();
				r_parser.LockParentNode();
				return false;
			}
			// put into the list
		} while(r_parser.LockNextNode());
		if(!r_parser.LockParentNode())
			return false;
	}
	// load menu pages

	if(!r_parser.LockParentNode())
		return false;

	_ASSERTE(r_parser.n_GetLastError() == Parser_Ok);

	return Navigate_FirstPage();
}

/*
 *	bool CPadsMenu::Navigate_FirstPage()
 *		- make sure the first (propably root) page is displayed
 */
bool CPadsMenu::Navigate_FirstPage()
{
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++) {
		if((*p_page)->n_Id() == m_n_first_page_id) {
			// todo - end editing action on current page if any
			m_p_cur_page = *p_page;
			return true;
		}
	}
	m_p_cur_page = 0;
	return false;
}

void CPadsMenu::Draw(CGLState *p_state)
{
	if(!m_p_cur_page)
		return;
	m_p_cur_page->Draw(p_state, m_p_stroke_font, m_t_metrics, m_t_color, m_n_button_num);
}

void CPadsMenu::Free()
{
	for(TToggleGroup *p_group = m_enum_list.begin(); p_group < m_enum_list.end(); p_group ++)
		p_group->Free();
	m_enum_list.clear();
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		(*p_page)->Free();
	m_page_list.clear();
	if(m_p_stroke_font) {
		delete m_p_stroke_font;
		m_p_stroke_font = 0;
	}
	m_p_cur_page = 0;
}

void CPadsMenu::OnMouseMove(Vector2f v_pos)
{
	if(!m_p_cur_page)
		return;
	int n = m_p_cur_page->n_OnMouseMove(v_pos, m_t_metrics, m_n_button_num);

	if(n != -1) {
		for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++) {
			if((*p_page)->n_Id() == n) {
				m_p_cur_page = *p_page;
				break;
			}
		}
	}
	// find target menu page
}

void CPadsMenu::OnMouseDown(Vector2f v_pos)
{
	if(!m_p_cur_page)
		return;
	int n = m_p_cur_page->n_OnMouseDown(v_pos, m_t_metrics, m_n_button_num);

	if(n != -1) {
		for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++) {
			if((*p_page)->n_Id() == n) {
				m_p_cur_page = *p_page;
				break;
			}
		}
	}
	// find target menu page
}

void CPadsMenu::OnMouseUp(Vector2f v_pos)
{
	if(!m_p_cur_page)
		return;
	int n = m_p_cur_page->n_OnMouseUp(v_pos, m_t_metrics, m_n_button_num);

	if(n != -1) {
		for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++) {
			if((*p_page)->n_Id() == n) {
				m_p_cur_page = *p_page;
				break;
			}
		}
	}
	// find target menu page
}

void CPadsMenu::OnKeyPress(int n_code)
{
	if(!m_p_cur_page)
		return;
	int n = m_p_cur_page->n_OnKeyPress(n_code, m_n_button_num);

	if(n != -1) {
		for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++) {
			if((*p_page)->n_Id() == n) {
				m_p_cur_page = *p_page;
				break;
			}
		}
	}
	// find target menu page
}

int CPadsMenu::n_CurButton_Token() const
{
	if(!m_p_cur_page)
		return 0;
	return m_p_cur_page->n_CurButton_Token();
}

int CPadsMenu::n_RegisterCallback(TCallbackHandler *p_callback_handler)
{
	int n_result = 0;
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		n_result += (*p_page)->n_RegisterCallback(p_callback_handler);
	return n_result;
}

void CPadsMenu::RemoveCallback(TCallbackHandler *p_callback_handler)
{
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		(*p_page)->RemoveCallback(p_callback_handler);
}

int CPadsMenu::n_AddWatch(const char *p_s_name, int *p_value_ptr)
{
	int n_result = 0;
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		n_result += (*p_page)->n_AddWatch(p_s_name, p_value_ptr);
	return n_result;
}

int CPadsMenu::n_AddWatch(const char *p_s_name, float *p_value_ptr)
{
	int n_result = 0;
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		n_result += (*p_page)->n_AddWatch(p_s_name, p_value_ptr);
	return n_result;
}

int CPadsMenu::n_AddWatch(const char *p_s_name, std::string *p_value_ptr)
{
	int n_result = 0;
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		n_result += (*p_page)->n_AddWatch(p_s_name, p_value_ptr);
	return n_result;
}

void CPadsMenu::RemoveWatch(const char *p_s_name)
{
	for(TMenuPage **p_page = m_page_list.begin(); p_page < m_page_list.end(); p_page ++)
		(*p_page)->RemoveWatch(p_s_name);
}

/*
 *								=== ~CPadsMenu ===
 */

/*
 *								=== CPadsMenu::TMenuMetrics ===
 */

bool CPadsMenu::TMenuMetrics::Load(/*const*/ CTextParser &r_parser)
{
	f_width = r_parser.f_Value("f_width", .3f, false);
	f_title_height = r_parser.f_Value("f_title_height", .05f, false);
	f_button_height = r_parser.f_Value("f_button_height", .05f, false);

	const std::vector<float> *p_vector;

	if(!(p_vector = r_parser.p_f_Array("v_button_spacing", -1, 0, false)) ||
	   p_vector->size() != 2)
		return false;
	v_button_spacing = Vector2f((*p_vector)[0], (*p_vector)[1]);

	if(!(p_vector = r_parser.p_f_Array("v_menu_offset", -1, 0, false)) ||
	   p_vector->size() != 2)
		return false;
	v_menu_offset = Vector2f((*p_vector)[0], (*p_vector)[1]);

	return r_parser.n_GetLastError() == Parser_Ok;
}

/*
 *								=== ~CPadsMenu::TMenuMetrics ===
 */

/*
 *								=== CPadsMenu::TMenuColor ===
 */

bool CPadsMenu::TMenuColor::Load(/*const*/ CTextParser &r_parser)
{
	const std::vector<float> *p_vector;

	if(!(p_vector = r_parser.p_f_Array("v_bk_color", -1, 0, false)) || p_vector->size() != 3)
		return false;
	v_bk_color = Vector3f((*p_vector)[0], (*p_vector)[1], (*p_vector)[2]);

	if(!(p_vector = r_parser.p_f_Array("v_focus_color", -1, 0, false)) || p_vector->size() != 3)
		return false;
	v_focus_color = Vector3f((*p_vector)[0], (*p_vector)[1], (*p_vector)[2]);

	if(!(p_vector = r_parser.p_f_Array("v_edit_color", -1, 0, false)) || p_vector->size() != 3)
		return false;
	v_edit_color = Vector3f((*p_vector)[0], (*p_vector)[1], (*p_vector)[2]);

	if(!(p_vector = r_parser.p_f_Array("v_border_color", -1, 0, false)) || p_vector->size() != 3)
		return false;
	v_border_color = Vector3f((*p_vector)[0], (*p_vector)[1], (*p_vector)[2]);

	if(!(p_vector = r_parser.p_f_Array("v_font_color", -1, 0, false)) || p_vector->size() != 3)
		return false;
	v_font_color = Vector3f((*p_vector)[0], (*p_vector)[1], (*p_vector)[2]);

	f_alpha = r_parser.f_Value("f_alpha", .7f, false);

	return r_parser.n_GetLastError() == Parser_Ok;
}

/*
 *								=== ~CPadsMenu::TMenuColor ===
 */

/*
 *								=== CSlider ===
 */

CSlider::CSlider(float *p_ref_value, float f_min, float f_max, float f_step, float f_snap)
	:m_p_ref_value(p_ref_value), m_f_min(f_min), m_f_max(f_max),
	m_f_step(f_step), m_f_snap(f_snap), m_b_snapped(false)
{
}

void CSlider::Draw(CGLState *p_state, Vector2f v_min,
	Vector2f v_max, const CPadsMenu::TMenuColor &r_t_menu_color)
{
	p_state->EnableBlend();
	p_state->BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	float f_hover_x = (*m_p_ref_value - m_f_min) / (m_f_max - m_f_min) *
		(v_max.x - v_min.x) + v_min.x;
	// calculate hover position

	_glDrawRect(p_state, v_min, Vector2f(f_hover_x, v_max.y), r_t_menu_color.v_focus_color,
		r_t_menu_color.f_alpha, GL_FILL);
	// fill - active

	_glDrawRect(p_state, Vector2f(f_hover_x, v_min.y), v_max, r_t_menu_color.v_bk_color,
		r_t_menu_color.f_alpha, GL_FILL);
	// fill - rest

	_glDrawRect(p_state, v_min, v_max, r_t_menu_color.v_border_color, 1.0f, GL_LINE);
	// border
}

void CSlider::OnMouseMove(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
	if(m_b_snapped) {
		float f_x = (v_pos.x - v_min.x) / (v_max.x - v_min.x);
		// calc x-coordinate

		if(f_x < .0f)
			f_x = .0f;
		if(f_x > 1.0f)
			f_x = 1.0f;
		// clamp it

		f_x *= m_f_max - m_f_min;
		f_x += m_f_min;
		// scale it

		if(m_f_step > .0f) {
			f_x -= m_f_min;
			f_x = floorf(f_x / m_f_step + .5f) * m_f_step;
			f_x += m_f_min;
		}
		// apply stepping if required

		*m_p_ref_value = f_x;
		// set output
	}
}

void CSlider::OnMouseDown(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
	if(m_b_snapped)
		m_b_snapped = false; // can happen if you release mouse button outside window

	if(v_pos.x < v_min.x || v_pos.x > v_max.x || v_pos.y < v_min.y || v_pos.y > v_max.y)
		return;
	// see if we hit the slider

	float f_hover_x = (*m_p_ref_value - m_f_min) / (m_f_max - m_f_min) *
		(v_max.x - v_min.x) + v_min.x;
	// calculate hover position

	if(fabsf(f_hover_x - v_pos.x) < m_f_snap) {
		m_b_snapped = true;
		//OnMouseMove(v_pos, v_min, v_max); // set value
	} else {
		float f_big_step = (m_f_step > .0f)? 100 * m_f_step : (m_f_max - m_f_min) * .125f;
		float f_x = *m_p_ref_value; // save to temp value, we don't want spikes in it
		
		f_x += ((f_hover_x < v_pos.x) * 2 - 1) * f_big_step;

		if(f_x < m_f_min)
			f_x = m_f_min;
		if(f_x > m_f_max)
			f_x = m_f_max;
		// clamp it

		if(m_f_step > .0f) {
			f_x -= m_f_min;
			f_x = floorf(f_x / m_f_step + .5f) * m_f_step;
			f_x += m_f_min;
		}
		// apply stepping if required

		*m_p_ref_value = f_x;
		// set output
	}
}

void CSlider::OnMouseUp(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
	m_b_snapped = false;
}

void CSlider::OnKeyPress(int n_code, char n_char)
{
	float f_big_step = (m_f_step > .0f)? 100 * m_f_step : (m_f_max - m_f_min) * .125f;
	float f_step = (m_f_step > .0f)? m_f_step : (m_f_max - m_f_min) * .0125f;
	float f_delta = .0f;

	switch(n_code) {
	case VK_LEFT:
	case VK_DOWN:
		f_delta = -f_step;
		break;
	case VK_RIGHT:
	case VK_UP:
		f_delta = f_step;
		break;
	case VK_PRIOR:
		f_delta = f_big_step;
		break;
	case VK_NEXT:
		f_delta = -f_big_step;
		break;
	case VK_HOME:
		f_delta = m_f_min - *m_p_ref_value;
		break;
	case VK_END:
		f_delta = m_f_max - *m_p_ref_value;
		break;
	}

	if(f_delta != .0f) {
		float f_x = *m_p_ref_value + f_delta;

		if(f_x < m_f_min)
			f_x = m_f_min;
		if(f_x > m_f_max)
			f_x = m_f_max;
		// clamp it

		if(m_f_step > .0f) {
			f_x -= m_f_min;
			f_x = floorf(f_x / m_f_step + .5f) * m_f_step;
			f_x += m_f_min;
		}
		// apply stepping if required

		*m_p_ref_value = f_x;
		// set output
	}
}

/*
 *								=== ~CSlider ===
 */

/*
 *								=== CEditBox ===
 */

/*class CEditBox {
protected:
	char *m_p_s_text;
	int m_n_alloc;
	int m_n_max_length;

	int m_n_cursor_pos; // !! cursor can be on both sides of selection
	int m_n_sel_start;
	int m_n_sel_length;

	CCharFilter *m_p_filter;

	float m_f_last_pos;

	bool m_b_dirty;

public:*/

CEditBox::CEditBox(const char *p_s_initial_text, CCharFilter *p_filter, int n_max_length)
	:m_p_filter(p_filter), m_n_max_length(n_max_length), m_b_dirty(false), m_f_last_pos(0)
{
	if(!p_s_initial_text) {
		m_p_s_text = 0;
		m_n_alloc = 0;

		m_n_cursor_pos = 0;
		m_n_sel_start = 0;
		m_n_sel_length = 0;
	} else {
		m_p_s_text = p_s_Clone(p_s_initial_text);
		m_n_alloc = (m_p_s_text)? strlen(m_p_s_text) + 1 : 0;

		if(n_max_length > 0 && n_max_length < m_n_alloc)
			m_p_s_text[n_max_length - 1] = 0;

		for(char *p_s_str = m_p_s_text; *p_s_str; p_s_str ++) {
			if(*p_s_str == '\n')
				*p_s_str = '/'; // no newlines
		}

		m_n_cursor_pos = m_n_alloc - 1;
		m_n_sel_start = 0;
		m_n_sel_length = m_n_alloc - 1;
	}
}

CEditBox::~CEditBox()
{
	if(m_p_s_text) {
		delete[] m_p_s_text;
		m_p_s_text = 0;
	}
}

void CEditBox::Draw(CGLState *p_state, const CStrokeFont *p_font, Vector2f v_min,
	Vector2f v_max, const CPadsMenu::TMenuColor &r_t_menu_color)
{
	p_state->EnableBlend();
	p_state->BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	_glDrawRect(p_state, v_min, v_max, r_t_menu_color.v_bk_color,
		r_t_menu_color.f_alpha, GL_FILL);
	// fill

	float f_char_height = -p_font->v_CharacterOffset("\n", 1).y;

	if(p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "") +
	   f_char_height * .5f <= v_max.x - v_min.x)
		m_f_last_pos = .0f; // i'd fit whole
	if(p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_cursor_pos) +
	   f_char_height * .5f + m_f_last_pos > v_max.x - v_min.x) {
		m_f_last_pos = -(p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_cursor_pos) +
			f_char_height * .5f - (v_max.x - v_min.x));
	} else if(p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_cursor_pos) -
	   f_char_height * .25f + m_f_last_pos < 0) {
		m_f_last_pos = -p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_cursor_pos);
	}
	// update position

	float f_cursor_height = f_char_height * 1.25f;
	float f_cursor_x = m_f_last_pos + v_min.x + f_char_height * .25f +
		p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_cursor_pos);

	float f_cursor_min_y = v_min.y - ((v_max.y - v_min.y) - f_cursor_height) * .5f;
	float f_cursor_max_y = v_max.y + ((v_max.y - v_min.y) - f_cursor_height) * .5f;

	Vector2f v_text_pos(m_f_last_pos + f_char_height * .25f,
		((v_max.y - v_min.y) + p_font->v_CharacterOffset("\n", 1).y) * .5f);
	v_text_pos += v_min;
	// calc text position

	float f_selection_x = m_f_last_pos + v_min.x + f_char_height * .25f +
		p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_sel_start);
	float f_cursor2_x = m_f_last_pos + v_min.x + f_char_height * .25f +
		p_font->f_StringWidth((m_p_s_text)? m_p_s_text : "", m_n_sel_start + m_n_sel_length);

	p_state->EnableClipPlane(0);
	p_state->EnableClipPlane(1);
	p_state->EnableClipPlane(2);
	p_state->EnableClipPlane(3);
	//
	p_state->ClipPlane(0, 1, 0, 0, -v_min.x);
	p_state->ClipPlane(1, 0, 1, 0, -v_min.y);
	p_state->ClipPlane(2, -1, 0, 0, v_max.x);
	p_state->ClipPlane(3, 0, -1, 0, v_max.y);
	// setup clip-planes

	_glDrawRect(p_state, Vector2f(f_selection_x, f_cursor_min_y),
		Vector2f(f_cursor2_x, f_cursor_max_y), r_t_menu_color.v_focus_color, 1.0f, GL_FILL);
	// draw selection

	p_state->Color4f(r_t_menu_color.v_font_color.x,
		r_t_menu_color.v_font_color.y, r_t_menu_color.v_font_color.z, 1.0f);
	p_font->Plot_String(p_state, v_text_pos, (m_p_s_text)? m_p_s_text : "");
	// draw text

	if((clock() / 512) & 1) {
		glBegin(GL_LINES);
		glVertex2f(f_cursor_x, f_cursor_min_y);
		glVertex2f(f_cursor_x, f_cursor_max_y);
		glEnd();
	}
	// draw cursor

	p_state->DisableClipPlane(0);
	p_state->DisableClipPlane(1);
	p_state->DisableClipPlane(2);
	p_state->DisableClipPlane(3);

	_glDrawRect(p_state, v_min, v_max, r_t_menu_color.v_border_color, 1.0f, GL_LINE);
	// border
}

bool CEditBox::b_Dirty()
{
	return m_b_dirty;
}

void CEditBox::ResetDirtyFlag()
{
	m_b_dirty = false;
}

const char *CEditBox::p_s_Text()
{
	return (m_p_s_text)? m_p_s_text : "";
}

void CEditBox::OnMouseMove(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
}

void CEditBox::OnMouseDown(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
}

void CEditBox::OnMouseUp(Vector2f v_pos, Vector2f v_min, Vector2f v_max)
{
}

void CEditBox::OnKeyPress(int n_code, char n_char)
{
	switch(n_code) {
	case VK_DELETE:
	case VK_BACK:
		if(m_n_sel_length)
			Delete_Selection();
		else {
			m_n_sel_length = 1;
			m_n_sel_start = m_n_cursor_pos - (n_code == VK_BACK);
			Delete_Selection();
		}
		break;
	case VK_END:
	case VK_HOME:
	case VK_LEFT:
	case VK_RIGHT:
		{
			bool b_shift = (GetAsyncKeyState(VK_SHIFT) & 0xff00) != 0;

			int n_prev_cursor_pos = m_n_cursor_pos;

			switch(n_code) {
			case VK_END:
				m_n_cursor_pos = (m_p_s_text)? strlen(m_p_s_text): 0;
				break;
			case VK_HOME:
				m_n_cursor_pos = 0;
				break;
			case VK_LEFT:
				m_n_cursor_pos --;
				break;
			case VK_RIGHT:
				m_n_cursor_pos ++;
				break;
			}

			if(m_n_cursor_pos < 0)
				m_n_cursor_pos = 0;
			if(m_n_cursor_pos > ((m_p_s_text)? strlen(m_p_s_text) : 0))
				m_n_cursor_pos = ((m_p_s_text)? strlen(m_p_s_text): 0);
			// calc new cursor pos

			if(b_shift) {
				if(n_prev_cursor_pos == m_n_sel_start) {
					if(m_n_sel_start + m_n_sel_length > m_n_cursor_pos) {
						m_n_sel_start = m_n_cursor_pos;
						m_n_sel_length += n_prev_cursor_pos - m_n_cursor_pos;
					} else {
						m_n_sel_start += m_n_sel_length;
						m_n_sel_length = m_n_cursor_pos - m_n_sel_start;
					}
				} else {
					if(m_n_sel_start > m_n_cursor_pos) {
						m_n_sel_length = m_n_sel_start - m_n_cursor_pos;
						m_n_sel_start = m_n_cursor_pos;
					} else
						m_n_sel_length = m_n_cursor_pos - m_n_sel_start;
				}
			} else {
				if(m_n_sel_length && n_code == VK_LEFT)
					m_n_cursor_pos = m_n_sel_start;
				else if(m_n_sel_length && n_code == VK_RIGHT)
					m_n_cursor_pos = m_n_sel_start + m_n_sel_length;
				m_n_sel_start = m_n_cursor_pos;
				m_n_sel_length = 0;
			}
			// update selection
		}
		break;
	default:
		if(n_char >= ' ') { // ignore lower chars
			if(m_p_filter && ((m_n_sel_length && !m_p_filter->b_Accept(n_char,
			   (m_p_s_text)? m_p_s_text : "", m_n_sel_start, m_n_sel_length)) ||
			   (!m_n_sel_length && !m_p_filter->b_Accept(n_char,
			   (m_p_s_text)? m_p_s_text : "", m_n_cursor_pos))))
				return;
			// see if character can pass trough filter

			if(m_n_sel_length) {
				m_b_dirty = true;
				Delete_Selection();
			}
			// we retype selected text

			if(!m_p_s_text || strlen(m_p_s_text) + 1 == m_n_alloc) {
				if(m_n_alloc == m_n_max_length)
					return;
				// we've already reached maximum charcount

				int n_alloc = m_n_alloc + 16;
				if(n_alloc > m_n_max_length && m_n_max_length > 0)
					n_alloc = m_n_max_length;
				char *p_s_new_str;
				if(!(p_s_new_str = new char[n_alloc]))
					return;
				strcpy(p_s_new_str, m_p_s_text);
				m_n_alloc = n_alloc;
				m_p_s_text = p_s_new_str;
				m_b_dirty = true;
			}
			// realloc if necessary and / or possible

			for(char *p_s_insert = m_p_s_text + m_n_cursor_pos,
			   *p_s_end = m_p_s_text + strlen(m_p_s_text) + 1; p_s_end > p_s_insert; p_s_end --)
				p_s_end[0] = p_s_end[-1];
			// make space for newly inserted char

			m_p_s_text[m_n_cursor_pos ++] = n_char;
			m_b_dirty = true;
		}
		break;
	}
}

/*
 *	void CEditBox::Delete_Selection()
 *		- delete selection
 *		- must handle such a cases as empty string or selection outside it, etc
 *		- update position of new cursor, selection, etc
 */
void CEditBox::Delete_Selection()
{
	if(!m_p_s_text || !*m_p_s_text) {
		m_n_cursor_pos = 0;
		m_n_sel_start = 0;
		m_n_sel_length = 0;

		return;
	}
	// empty string

	if(m_n_sel_start < 0) {
		m_n_sel_length += m_n_sel_start;
		m_n_sel_start = 0;
	}
	if(m_n_sel_length < 0)
		m_n_sel_length = 0;
	if(m_n_sel_start > strlen(m_p_s_text)) { // must be able to write past last character
		m_n_sel_start = strlen(m_p_s_text);
		m_n_sel_length = 0;
	}
	if(m_n_sel_start + m_n_sel_length > strlen(m_p_s_text)) // right, not '>=' ... m_n_sel_length is length, not index
		m_n_sel_length = strlen(m_p_s_text) - m_n_sel_start;
	// handle various errors

	_ASSERTE(m_n_sel_start >= 0);
	_ASSERTE(m_n_sel_start <= strlen(m_p_s_text));
	_ASSERTE(m_n_sel_length >= 0);
	_ASSERTE(m_n_sel_length <= strlen(m_p_s_text));
	_ASSERTE(m_n_sel_start + m_n_sel_length >= 0);
	_ASSERTE(m_n_sel_start + m_n_sel_length <= strlen(m_p_s_text));

	if(m_n_sel_length) {
		for(char *p_s_dst = m_p_s_text + m_n_sel_start,
		   *p_s_src = m_p_s_text + m_n_sel_start + m_n_sel_length;; p_s_src ++) {
			*p_s_dst ++ = *p_s_src;
			if(!*p_s_src)
				break;
			// want to copy zero as well
		}
		m_n_sel_length = 0;
	}
	// erase if there is something to erase

	m_b_dirty = true;

	m_n_cursor_pos = m_n_sel_start;
}

/*
 *								=== ~CEditBox ===
 */

/*
 *								===  ===
 */



/*
 *								=== ~ ===
 */
