/*
								+---------------------------------+
								|                                 |
								|      ***   OpenGL2.0   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|           Shader2.inl           |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 10-09-2006
 *
 *	removed most of berLame stuff (todo - write new shader manager and generic encapsulation class)
 *	integers, containing just true or false were changed to bool
 *	fixed minor flaws in shader binding
 *	renamed to Shader2.cpp and Shader2.h respectively to avoid confusion with similar old source
 *
 *	passed code revision at 11-10-2006
 *
 *	created inline versions of Bind, Release and SetParam* functions so the encapsulation class,
 *	aware of shader api could use them (virtual functions themselves call them so there is no
 *	code written twice); note inlines are in Shader2.inl, there is too much of them to put them here
 *
 */

#ifndef __GL_SHADER2_INLINES_INCLUDED
#define __GL_SHADER2_INLINES_INCLUDED

/*
 *								=== CGL_ARB_program ===
 */

__inline bool CGL_ARB_program::_SetParam4fv(CGLState *p_state, int n_param_type,
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_ARB_program::_SetParam4fv");

	if(n_param_type == shader_param_Parameter_Env) {
		if(m_p_s_vertex_shader)
			glProgramEnvParameter4fvARB(GL_VERTEX_PROGRAM_ARB, n_location, p_value);
		else if(m_p_s_fragment_shader)
			glProgramEnvParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, n_location, p_value);
		else
			return false; // no shader available
	} else if(n_param_type == shader_param_Parameter_Local_VP) {
		if(m_p_s_vertex_shader)
			glProgramLocalParameter4fvARB(GL_VERTEX_PROGRAM_ARB, n_location, p_value);
		else
			return false; // bad shader type
	} else if(n_param_type == shader_param_Parameter_Local_FP) {
		if(m_p_s_vertex_shader)
			glProgramLocalParameter4fvARB(GL_FRAGMENT_PROGRAM_ARB, n_location, p_value);
		else
			return false; // bad shader type
	} else
		return false; // bad param type

	return true;
}

__inline void CGL_ARB_program::_Bind(CGLState *p_state) const
{
	__FuncGuard("CGL_ARB_program::_Bind");

	//CGLProgram::Bind(p_state);

	if(m_n_vs_object) {
		p_state->EnableVertexProgram();
		p_state->BindVertexProgram(m_n_vs_object);
	} else
		p_state->DisableVertexProgram();
	if(m_n_fs_object) {
		p_state->EnableFragmentProgram();
		p_state->BindFragmentProgram(m_n_fs_object);
	} else
		p_state->DisableFragmentProgram();
}

__inline void CGL_ARB_program::_Release(CGLState *p_state) const
{
	__FuncGuard("CGL_ARB_program::_Release");

	//CGLProgram::Release(p_state);

	p_state->DisableVertexProgram();
	p_state->DisableFragmentProgram();
}

/*
 *								=== ~CGL_ARB_program ===
 */

/*
 *								=== CGL_ARB_shader ===
 */

__inline bool CGL_ARB_shader::_SetParam1fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam1fv");

	glUniform1fvARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam2fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam2fv");

	glUniform2fvARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam3fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam3fv");

	glUniform3fvARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam4fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam4fv");

	glUniform4fvARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam1i(CGLState *p_state,
	int n_location, int n_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam1i");

	glUniform1iARB(n_location, n_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam2iv(CGLState *p_state, int n_location,
	const int *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam2iv");

	glUniform2ivARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam3iv(CGLState *p_state, int n_location,
	const int *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam3iv");

	glUniform3ivARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParam4iv(CGLState *p_state, int n_location,
	const int *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParam4iv");

	glUniform4ivARB(n_location, 1, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParamMatrix2fv(CGLState *p_state, int n_location,
	bool b_transpose, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParamMatrix2fv");

	glUniformMatrix2fvARB(n_location, 1, b_transpose, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParamMatrix3fv(CGLState *p_state, int n_location,
	bool b_transpose, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParamMatrix3fv");

	glUniformMatrix3fvARB(n_location, 1, b_transpose, p_value);

	return true;
}

__inline bool CGL_ARB_shader::_SetParamMatrix4fv(CGLState *p_state, int n_location,
	bool b_transpose, const float *p_value) const
{
	__FuncGuard("CGL_ARB_shader::_SetParamMatrix4fv");

	glUniformMatrix4fvARB(n_location, 1, b_transpose, p_value);

	return true;
}

__inline void CGL_ARB_shader::_Bind(CGLState *p_state) const
{
	__FuncGuard("CGL_ARB_shader::_Bind");

	//CGLProgram::Bind(p_state);

	p_state->BindProgramObject(m_n_program_object);
}

__inline void CGL_ARB_shader::_Release(CGLState *p_state) const
{
	__FuncGuard("CGL_ARB_shader::_Release");

	//CGLProgram::Release(p_state);

	p_state->BindProgramObject(0);
}

/*
 *								=== ~CGL_ARB_shader ===
 */

/*
 *								=== CGL_Core_shader ===
 */

__inline bool CGL_Core_shader::_SetParam1fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam1fv");

	glUniform1fv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam2fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam2fv");

	glUniform2fv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam3fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam3fv");

	glUniform3fv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam4fv(CGLState *p_state, /*int n_param_type,*/
	int n_location, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam4fv");

	glUniform4fv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam1i(CGLState *p_state, int n_location, int n_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam1i");

	glUniform1i(n_location, n_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam2iv(CGLState *p_state, int n_location,
	const int *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam2iv");

	glUniform2iv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam3iv(CGLState *p_state, int n_location,
	const int *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam3iv");

	glUniform3iv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParam4iv(CGLState *p_state, int n_location,
	const int *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParam4iv");

	glUniform4iv(n_location, 1, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParamMatrix2fv(CGLState *p_state, int n_location,
	bool b_transpose, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParamMatrix2fv");

	glUniformMatrix2fv(n_location, 1, b_transpose, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParamMatrix3fv(CGLState *p_state, int n_location,
	bool b_transpose, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParamMatrix3fv");

	glUniformMatrix3fv(n_location, 1, b_transpose, p_value);

	return true;
}

__inline bool CGL_Core_shader::_SetParamMatrix4fv(CGLState *p_state, int n_location,
	bool b_transpose, const float *p_value) const
{
	__FuncGuard("CGL_Core_shader::_SetParamMatrix4fv");

	glUniformMatrix4fv(n_location, 1, b_transpose, p_value);

	return true;
}

__inline void CGL_Core_shader::_Bind(CGLState *p_state) const
{
	__FuncGuard("CGL_Core_shader::_Bind");

	//CGLProgram::Bind(p_state);

	p_state->BindProgramObject_GL2(m_n_program_object);
}

__inline void CGL_Core_shader::_Release(CGLState *p_state) const
{
	__FuncGuard("CGL_Core_shader::_Release");

	//CGLProgram::Release(p_state);

	p_state->BindProgramObject_GL2(0);
}

/*
 *								=== ~CGL_Core_shader ===
 */

#endif //__GL_SHADER2_INLINES_INCLUDED
