/*
								+----------------------------------+
								|                                  |
								|   ***  Stroke font-engine  ***   |
								|                                  |
								|   Copyright  -tHE SWINe- 2005   |
								|                                  |
								|          StrokeFont.cpp          |
								|                                  |
								+----------------------------------+
*/

#include <vector>
#include "CallStack.h"
#include <windows.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <gl/gl.h>
#include <gl/glext.h>
#include "Pashik_TextParser2.h"
#include "Vector.h"
#include "OpenGLState.h"
#include "StrokeFont.h"

/*
 *								=== TStrokeChar ===
 */

struct TStrokeChar {
	TStrokeChar() :p_strip_len(0), p_strip(0) { }
	~TStrokeChar() { if(p_strip_len) delete[] p_strip_len; p_strip_len = 0;
					 if(p_strip) delete[] p_strip; p_strip = 0; }
	//
	char n_ascii;
	//
	int n_strip_num;
	int *p_strip_len;
	Vector2f *p_strip;
	//
	Vector2f v_org;
	float f_width;
};

/*
 *								=== ~TStrokeChar ===
 */

/*
 *								=== CStrokeFont ===
 */

CStrokeFont::CStrokeFont()
	:m_p_char(0)
{
}

CStrokeFont::~CStrokeFont()
{
	Free();
}

int CStrokeFont::Load(const char *p_s_filename, CGLState *p_state)
{
	Free();

	CTextParser parser;

	if(!parser.Attach(p_s_filename) || !parser.LoadFromFile())
		return false;

	// parse file

	m_n_char_num = parser.n_Value("n_glyph_num", 0, false);
	m_b_space_style = parser.n_Value("b_space_style", 0, false);
	m_f_space_shift = parser.f_Value("f_space_shift", 2.0f, false);
	m_f_grid_height = parser.f_Value("f_grid_height", 5.0f, false);
	m_f_cur_size = 1.0f;
	m_f_side_ratio = parser.f_Value("f_side_ratio", 1.0f, false);
	//
	if(parser.n_GetLastError() != Parser_Ok)
		return false;
	// read font config

	if(!(m_p_char = new TStrokeChar[m_n_char_num]))
		return false;
	// alloc char memory

	for(int i = 0; i < m_n_char_num; i ++) {
		char p_s_key[32];

		sprintf(p_s_key, "glyph[%d].n_ascii", i);
		m_p_char[i].n_ascii = parser.n_Value(p_s_key, 0, false);

		sprintf(p_s_key, "glyph[%d].p_data", i);
		const std::vector<int> *p_array = parser.p_n_Array(p_s_key, 0, 0, false);
		if(parser.n_GetLastError() != Parser_Ok || !p_array)
			return false;

		sprintf(p_s_key, "glyph[%d].n_org_x", i);
		m_p_char[i].v_org.x = (float)parser.n_Value(p_s_key, 0, false);
		sprintf(p_s_key, "glyph[%d].n_org_y", i);
		m_p_char[i].v_org.y = (float)parser.n_Value(p_s_key, 0, false);
		parser.n_GetLastError();
		// org-x & y are voluntary

		int n_data_size = 0;
		m_p_char[i].n_strip_num = 0;
		for(int j = 0; j < p_array->size(); j += (*p_array)[j] * 2 + 1) {
			m_p_char[i].n_strip_num ++;
			n_data_size += (*p_array)[j];
		}
		// count strips (array contains <strip_len> <strip_data> ...)

		if(!(m_p_char[i].p_strip_len = new int[m_p_char[i].n_strip_num]))
			return false;
		if(!(m_p_char[i].p_strip = new Vector2f[n_data_size])) {
			delete m_p_char[i].p_strip_len;
			return false;
		}
		for(int j = 0, n = 0, k = 0; j < p_array->size(); /*j ++, */n ++) {
			_ASSERTE(n < m_p_char[i].n_strip_num);
			m_p_char[i].p_strip_len[n] = (*p_array)[j];
			if(!m_p_char[i].p_strip_len[n] ||
			   m_p_char[i].p_strip_len[n] * 2 + j >= p_array->size())
				return false; // malformed data
			int n_last = ++ j + m_p_char[i].p_strip_len[n] * 2;
			for(; j < n_last; j += 2, k ++) {
				_ASSERTE(k < n_data_size);
				m_p_char[i].p_strip[k] = Vector2f((*p_array)[j], m_f_grid_height-(*p_array)[j + 1]);
			}
		}
		// read strips
	}
	// read char data

	for(int i = 0; i < m_n_char_num; i ++) {
		float f_max = -65536.0f, f_min = 65536.0f;
		for(int j = 0, k = 0, l = 0;; l ++, j ++) {
			if(l == m_p_char[i].p_strip_len[k]) {
				l = 0;
				if(++ k == m_p_char[i].n_strip_num)
					break;
			}

			if(m_p_char[i].p_strip[j].x > f_max)
				f_max = m_p_char[i].p_strip[j].x;
			if(m_p_char[i].p_strip[j].x < f_min)
				f_min = m_p_char[i].p_strip[j].x;
		}

		m_p_char[i].f_width = f_max - f_min;
	}
	m_f_grid_height += 1;
	// calc char width

	for(int j = 0; j < 256; j ++) {
		m_n_lookup[j] = -1;
		for(int i = 0; i < m_n_char_num; i ++) {
			if((unsigned char)m_p_char[i].n_ascii == j) {
				m_n_lookup[j] = i;
				break;
			}
		}
	}
	// hash table

	return true;
}

void CStrokeFont::Free()
{
	if(m_p_char) {
		delete[] m_p_char;
		m_p_char = 0;
	}
}

void CStrokeFont::Zoom(float f_new_size)
{
	m_f_cur_size = f_new_size;
}

void CStrokeFont::Plot_Character(CGLState *p_state, Vector2f v_pos, char c_char) const
{
	if(!m_p_char)
		return;

	if(m_n_lookup[(unsigned char)c_char] == -1)
		return;
	if(c_char == ' ') {
		if(m_b_space_style)
			c_char = '_';
		else
			return;
	}
	TStrokeChar *p_char = &m_p_char[m_n_lookup[(unsigned char)c_char]];
	// find appropriate character

	v_pos -= p_char->v_org * m_f_cur_size;
	for(int i = 0, n = 0; i < p_char->n_strip_num; i ++) {
		glBegin(GL_LINE_STRIP);
		for(int j = 0; j < p_char->p_strip_len[i]; j ++, n ++) {
			glVertex2f((p_char->p_strip[n] * m_f_cur_size + v_pos).x,
					   (p_char->p_strip[n] * m_f_cur_size * m_f_side_ratio + v_pos).y);
		}
		glEnd();
	}
	// draw a character
}

void CStrokeFont::Plot_String(CGLState *p_state, Vector2f v_pos,
	const char *p_s_string, int n_length) const
{
	if(!m_p_char)
		return;

	if(n_length == -1)
		n_length = strlen(p_s_string);

	float f = 0, v = 0;
	for(int i = 0; i < n_length; i ++) {
		if(p_s_string[i] == '\n') {
			f = 0;
			v -= m_f_grid_height * m_f_cur_size * m_f_side_ratio;
			continue;
		}

		Plot_Character(p_state, v_pos + Vector2f(f, v), p_s_string[i]);

		if(m_n_lookup[(unsigned char)p_s_string[i]] == -1)
			f += m_f_space_shift * m_f_cur_size;
		else
			f += (m_p_char[m_n_lookup[(unsigned char)p_s_string[i]]].f_width + 1) * m_f_cur_size;
	}
}

float CStrokeFont::f_StringWidth(const char *p_s_string, int n_length) const
{
	if(n_length == -1)
		n_length = strlen(p_s_string);

	float f = 0;
	for(int i = 0; i < n_length; i ++) {
		if(p_s_string[i] == '\n')
			return f * m_f_cur_size;

		if(m_n_lookup[(unsigned char)p_s_string[i]] != -1) {
			f += m_p_char[m_n_lookup[(unsigned char)p_s_string[i]]].f_width + 1;
		} else
			f += m_f_space_shift;
	}

	return f * m_f_cur_size;
}

Vector2f CStrokeFont::v_CharacterOffset(const char *p_s_string, int n_length) const
{
	if(n_length == -1)
		n_length = strlen(p_s_string);

	Vector2f v(0, 0);
	for(int i = 0; i < n_length; i ++) {
		if(p_s_string[i] == '\n') {
			v.y -= m_f_grid_height * m_f_side_ratio;
			v.x = 0;
			continue;
		}

		if(m_n_lookup[(unsigned char)p_s_string[i]] != -1) {
			v.x += m_p_char[m_n_lookup[(unsigned char)p_s_string[i]]].f_width + 1;
		} else
			v.x += m_f_space_shift;
	}

	return v * m_f_cur_size;
}

/*
 *								=== ~CStrokeFont ===
 */
